
//
// OS.c
//
// <see corresponding .h file>
// 
//
// $Id: OS.c,v 1.6 2001/04/23 23:51:47 dru Exp $
//
//
//
//

#include "cel.h"
#include "runtime.h"
#include "svm32.h"
#include "array.h"

#include <unistd.h>
#include <sys/utsname.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <string.h>


//
//  OS - this is the wrapper over Unix for Cel
// 

extern char *slotList[];
char *slotList[] = {
	"getProcessId", "OSgetProcessId",
	"getGroupId", "OSgetGroupId",
	"getUserId", "OSgetUserId",
	"fork",      "OSfork",
	"exec:",      "OSexec",
	"wait",      "OSwait",
//	"getCurrentDir", "OSgetCurrentDir",
	"waitPid:", "OSwaitPid",
	"sleep:", "OSsleep",
	"openForWrite:", "OSopenForWrite",
	"openForRead:", "OSopenForRead",
	"close:", "OSclose",
	"readFd:Into:Size:", "OSreadFdIntoSize",
	"writeFd:From:", "OSwriteFdFrom",
// Notice use of same method
	"exclusiveLockFd:", "OSflock",
	"unLockFd:", "OSflock",
	"exit:", "OSexit",
	"getUname", "OSgetUname",
	"getHostname", "OSgetHostname",
	"createPipe", "OScreatePipe",
	"dupFd:to:", "OSdup2",
	"unlink:", "OSunlink",
	"rename:To:", "OSrenameTo",
	"system:", "OSsystem",
	""
};

extern char **environ;


void OSgetProcessId(void)
{
    Proto p;

    p = objectNewInteger(getpid());
    VMReturn(Cpu, (unsigned int) p, 3);
}

void OSgetGroupId(void)
{
    Proto p;

    p = objectNewInteger(getgid());
    VMReturn(Cpu, (unsigned int) p, 3);
}

void OSgetUserId(void)
{
    Proto p;

    p = objectNewInteger(getuid());
    VMReturn(Cpu, (unsigned int) p, 3);
}

void OSfork(void)
{
    Proto p;

    p = objectNewInteger(fork());
    VMReturn(Cpu, (unsigned int) p, 3);
}

void OSexec(void)
{
    Proto target;
    Proto blob;
    Proto tmpObj;
    char buff[128];
    int size;
    int i;
    char * p;
    char * tp;
    char ** ap;
    char ** argv;
    char * args[2];
    

    target    = (Proto) stackAt(Cpu,4);

    tp = NULL;
    size = 0;
    // Normalize the target
    i = PROTO_REFTYPE(target);
    if (i == PROTO_REF_ATOM) {
	tp = atomToString((Atom)target);
	size = strlen(tp);
    }
    else if (i == PROTO_REF_PROTO) {
	i = objectGetSlot(target, stringToAtom("_string"), &blob);
	if (i == 0) {
	    tp = "";
	    size = 0;
	} else {
	    tp = objectPointerValue(blob);
	    assert(objectGetSlot(target, stringToAtom("size"), &tmpObj));
	    size = objectIntegerValue(tmpObj);
	}
    }

    // Make a copy of the command
    assert(size < 126);
    memcpy(buff, tp, size);
    buff[size] = 0;
    tp = buff;


    // My old hack was to just run a '/bin/sh -c ' and let
    // it parse the command line. Nowadays, I will parse the thing
    p = strchr(tp, ' ');
    if (p == NULL) {
	args[0] = tp;
	args[1] = "";
	execve(tp, args, environ);
    } else {
	argv = celcalloc((sizeof(char *)) * (25));
	for (ap = argv; (*ap = strsep(&tp, " \t")) != NULL;) {
	    if (**ap != '\0') {
		if (++ap >= &argv[24]) {
		    break;
		}
	    }
	}
	execve(buff, argv, environ);
    }
	
    // We shouldn't get here

}

void OSwait(void)
{
    Proto p;
    Proto statObj;
    int status;

    p = objectNewInteger(wait(&status));

    statObj = createArray(nil);
    addToArray(statObj, p);
    
    addToArray(statObj, objectNewInteger(WEXITSTATUS(status)));
    
    VMReturn(Cpu, (unsigned int) statObj, 3);
}

void OSwaitPid(void)
{
    Proto p;
    Proto arg;
    Proto statObj;
    int status;
    int i;

    arg = (Proto) stackAt(Cpu,4);

    i = waitpid(objectIntegerValue(arg), &status, 0);
    
    p = objectNewInteger(i);

    statObj = createArray(nil);
    addToArray(statObj, p);
    
    addToArray(statObj, objectNewInteger(WEXITSTATUS(status)));
    
    VMReturn(Cpu, (unsigned int) statObj, 4);
}

void OSsleep(void)
{
    Proto p;
    Proto arg;

    arg    = (Proto) stackAt(Cpu,4);
    p = objectNewInteger(sleep(objectIntegerValue(arg)));
    VMReturn(Cpu, (unsigned int) p, 4);
}

void OSopenForWrite (void)
{
    Proto p;
    Proto arg;
    int fd;
    int i;
    char * ap;
    int  argSize;
    Proto argBlob;
    Proto tmpObj;
    

    arg    = (Proto) stackAt(Cpu,4);

    // Normalize the argument
    i = PROTO_REFTYPE(arg);
    if (i == PROTO_REF_ATOM) {
	ap = atomToString((Atom) arg);
	argSize = strlen(ap);
    }
    else if (i == PROTO_REF_PROTO) {
	assert(objectGetSlot(arg, stringToAtom("_string"), &argBlob));
	ap = objectPointerValue(argBlob);
	assert(objectGetSlot(arg, stringToAtom("size"), &tmpObj));
	argSize = objectIntegerValue(tmpObj);
    }

    fd = open(ap, O_WRONLY|O_CREAT|O_TRUNC, 0770);
    
    p = objectNewInteger(fd);
    VMReturn(Cpu, (unsigned int) p, 4);
}

void OSopenForRead (void)
{
    Proto p;
    Proto arg;
    int fd;
    int i;
    char * ap;
    int  argSize;
    Proto argBlob;
    Proto tmpObj;
    

    arg    = (Proto) stackAt(Cpu,4);

    // Normalize the argument
    i = PROTO_REFTYPE(arg);
    if (i == PROTO_REF_ATOM) {
	ap = atomToString((Atom) arg);
	argSize = strlen(ap);
    }
    else if (i == PROTO_REF_PROTO) {
	assert(objectGetSlot(arg, stringToAtom("_string"), &argBlob));
	ap = objectPointerValue(argBlob);
	assert(objectGetSlot(arg, stringToAtom("size"), &tmpObj));
	argSize = objectIntegerValue(tmpObj);
    }

    fd = open(ap, O_RDONLY);
    
    p = objectNewInteger(fd);
    VMReturn(Cpu, (unsigned int) p, 4);
}

void OSclose(void)
{
    Proto p;
    Proto arg;

    arg    = (Proto) stackAt(Cpu,4);
    p = objectNewInteger(close(objectIntegerValue(arg)));
    VMReturn(Cpu, (unsigned int) p, 4);
}

void OSexit(void)
{
    Proto arg;

    arg    = (Proto) stackAt(Cpu,4);
    exit(objectIntegerValue(arg));
    // As if :-)
    VMReturn(Cpu, (unsigned int) arg, 3);
}

void OSgetUname(void)
{
    Proto p;
    struct utsname name;

    if (uname(&name) == 0) {
	p = objectNew(0);
	objectSetSlot(p, stringToAtom("sysname"), PROTO_ACTION_GET, (uint32) stringToAtom(name.sysname));
	objectSetSlot(p, stringToAtom("nodename"), PROTO_ACTION_GET, (uint32) stringToAtom(name.nodename));
	objectSetSlot(p, stringToAtom("release"), PROTO_ACTION_GET, (uint32) stringToAtom(name.release));
	objectSetSlot(p, stringToAtom("version"), PROTO_ACTION_GET, (uint32) stringToAtom(name.version));
	objectSetSlot(p, stringToAtom("machine"), PROTO_ACTION_GET, (uint32) stringToAtom(name.machine));
	
    } else {
	p = nil;
    }
    
    VMReturn(Cpu, (unsigned int) p, 3);
}

void OSgetHostname(void)
{
    Atom p;
    char buff[64];
    
    if (gethostname(buff, sizeof(buff)) == 0) {
	p = stringToAtom(buff);
    } else {
	p = nil;
    }
    
    VMReturn(Cpu, (unsigned int) p, 3);
}

void OScreatePipe(void)
{
    int fds[2];
    Proto array;
    int i;
    
    
    i = pipe(fds);

    if (i != 0) {
	VMReturn(Cpu, (unsigned int) nil, 3);
	return;
    }
    
    array = createArray(nil);

    addToArray(array, objectNewInteger(fds[0]));
    addToArray(array, objectNewInteger(fds[1]));

    VMReturn(Cpu, (unsigned int) array, 3);
}

void OSdup2(void)
{
    Proto target;
    Proto oldFdObj;
    Proto newFdObj;
    int i;
    int new, old;
    
    
    target      = (Proto) stackAt(Cpu,2);
    oldFdObj    = (Proto) stackAt(Cpu,4);
    newFdObj    = (Proto) stackAt(Cpu,5);

    old = objectIntegerValue(oldFdObj);
    new = objectIntegerValue(newFdObj);

    i = dup2(old, new);
    
    if (i == -1) {
	VMReturn(Cpu, (unsigned int) nil, 5);
	return;
    }
    
    VMReturn(Cpu, (unsigned int) target, 5);
}

void OSunlink(void)
{
    Proto target;
    int i;
    Proto obj;
    char tbuff[128];

    target    = (Proto) stackAt(Cpu,4);

    getString(target, tbuff, sizeof(tbuff));
    
    i = unlink(tbuff);
    
    obj = objectNewInteger(i);
    VMReturn(Cpu, (unsigned int) obj, 4);
}

void OSrenameTo(void)
{
    Proto target;
    Proto obj;
    int i;
    Proto arg;
    char  tbuff[128];
    char  abuff[128];

    target    = (Proto) stackAt(Cpu,4);
    arg       = (Proto) stackAt(Cpu,5);

    getString(target, tbuff, sizeof(tbuff));
    getString(arg,    abuff, sizeof(abuff));

    i = rename(tbuff, abuff);

    obj = objectNewInteger(i);
    VMReturn(Cpu, (unsigned int) obj, 5);
}



void OSreadFdIntoSize(void)
{
    Proto buffObj;
    Proto sizeObj;
    Proto tmpObj;
    Proto fdObj;
    char * buff;
    int fd;
    int size;
    int stringSize;
    Proto self;
    int i;
    

    self    = (Proto) stackAt(Cpu,2);
    fdObj   = (Proto) stackAt(Cpu,4);
    buffObj = (Proto) stackAt(Cpu,5);
    sizeObj = (Proto) stackAt(Cpu,6);

    fd = objectIntegerValue(fdObj);


    if (objectGetSlot(buffObj, stringToAtom("_string"), &tmpObj) ){
	buff = objectPointerValue(tmpObj);
	assert(objectGetSlot(buffObj, stringToAtom("capacity"), &tmpObj) );
	stringSize = objectIntegerValue(tmpObj);
    } else {
	VMReturn(Cpu, (unsigned int) nil, 6);
    }
    
    if (PROTO_REFTYPE(sizeObj) == PROTO_REF_INT) {
	size = objectIntegerValue(sizeObj);
    } else {
	VMReturn(Cpu, (unsigned int) nil, 6);
    }

    if (size > stringSize) {
	VMReturn(Cpu, (unsigned int) nil, 6);
    }
    
    i = read(fd, buff, size);

    // Set the new size if no error
    if (i > 0) {
	objectSetSlot(buffObj, stringToAtom("size"), PROTO_ACTION_GET, (uint32) objectNewInteger(i));
    }
    

    VMReturn(Cpu, (unsigned int) objectNewInteger(i), 6);
}


void OSwriteFdFrom (void)
{
    Proto buffObj;
    Proto tmpObj;
    Proto fdObj;
    Proto self;
    char * buff;
    int fd;
    int stringSize;
    int i;
    
    self    = (Proto) stackAt(Cpu,2);
    fdObj   = (Proto) stackAt(Cpu,4);
    buffObj = (Proto) stackAt(Cpu,5);

    fd = objectIntegerValue(fdObj);

    if (PROTO_REFTYPE(buffObj) == PROTO_REF_ATOM) {
	buff = objectPointerValue(buffObj);
	stringSize = strlen(buff);
    }
    else
    if (objectGetSlot(buffObj, stringToAtom("_string"), &tmpObj) ){
	buff = objectPointerValue(tmpObj);
	assert(objectGetSlot(buffObj, stringToAtom("size"), &tmpObj) );
	stringSize = objectIntegerValue(tmpObj);
    } else {
	VMReturn(Cpu, (unsigned int) nil, 5);
    }
    
    i = write(fd, buff, stringSize);

    VMReturn(Cpu, (unsigned int) objectNewInteger(i), 5);
}


void OSflock (void)
{
    Atom mesg;
    Proto fdObj;
    Proto self;
    int fd;
    int i;
    struct flock lockStruct;
    
    
    self    = (Proto) stackAt(Cpu,2);
    mesg    = (Atom) stackAt(Cpu,3);
    fdObj   = (Proto) stackAt(Cpu,4);

    fd = objectIntegerValue(fdObj);

    lockStruct.l_whence = 0;
    lockStruct.l_start  = 0;
    lockStruct.l_len    = 0;
    lockStruct.l_pid    = 0;

    if (mesg == stringToAtom("exclusiveLockFd:")) {
	lockStruct.l_type = F_WRLCK;
	i = fcntl(fd, F_SETLKW, &lockStruct);
    }
    else if (mesg == stringToAtom("unLockFd:")) {
	lockStruct.l_type = F_UNLCK;
	i = fcntl(fd, F_SETLKW, &lockStruct);
    }

    VMReturn(Cpu, (unsigned int) objectNewInteger(i), 4);
}


void OSsystem(void)
{
    Proto target;
    Proto obj;
    int i;
    char tbuff[128];
    
    target    = (Proto) stackAt(Cpu,4);

    getString(target, tbuff, sizeof(tbuff));
    
    i = system(tbuff);
    
    obj = objectNewInteger(i);
    VMReturn(Cpu, (unsigned int) obj, 4);
}

