//
// svm32.h
//
// This is the description of the functions for the Stack-VM 32 bit
// 
//
// $Id: svm32.h,v 1.9 2001/03/07 06:56:35 dru Exp $
//
//
//
//


#ifndef H_CELSVM32
#define H_CELSVM32


#define CC_CARRY 0x01
#define CC_ZERO  0x02
#define CC_SIGN  0x04
#define CC_OF	 0x08

struct _vm32cpu {
	unsigned int pc;	/* Program Counter */
	unsigned int cc;	/* Condition Code */
	unsigned int sp;	/* Stack Pointer */
	unsigned int fp;	/* Frame Pointer */
	unsigned int haltflag;	/* Halt Flag :-) */
	// The following two are for branches
	unsigned int trueObj;	/* Value that references true obj */
	unsigned int falseObj;	/* Value that references false obj */
	unsigned int stackBottom;	/* Stack bottom, if we go past this, we're toast */
	unsigned int stackTop;          /* Stack top, if we go past this, we're toast */
	void (*breakHook) (struct _vm32cpu *);	/* Break Handler */
	void (*invalidHook)(struct _vm32cpu *);	/* Invalid Instruction Handler */
	void (*argCountHook)(struct _vm32cpu *);	/* Invalid message argument count hook */
	unsigned int (*pushHook)(struct _vm32cpu *, unsigned int);	/* Hook for special pushImmediate handling */
};


typedef struct _vm32cpu VM32Cpu;


inline void push(VM32Cpu * cpu, unsigned int v);
inline unsigned int pull(VM32Cpu * cpu);
inline unsigned int stackAt(VM32Cpu * cpu, unsigned int i);
void setStackAt(VM32Cpu * cpu, unsigned int i, unsigned int data);

inline unsigned int getFrameAt(VM32Cpu * cpu, int i);
inline unsigned int getFPRelativeAt(unsigned int fp, int i);

inline unsigned int setFrameAt(VM32Cpu * cpu, int pos, unsigned int value);
inline unsigned int setFPRelativeAt(unsigned int fp, int pos, unsigned int value);


void init (VM32Cpu * cpu);

void reset (VM32Cpu * );
int run (VM32Cpu * );
int singleStep (VM32Cpu * );
int halt (VM32Cpu *);


// User level routines 
inline void VMReturn(VM32Cpu * cpu, unsigned int returnValue, unsigned int moveCount);

// Debugger Support
int disassembleCpu(VM32Cpu * cpu, char * * opcode,  int * length);


#endif
